@@ -14,9 +14,11 @@
 
 import asyncio
 from asyncio.subprocess import DEVNULL, PIPE, STDOUT
+import codecs
 from collections import OrderedDict
 import io
 import os
+from pathlib import PurePath
 import socket
 import stat
 
@@ -39,41 +41,28 @@ class _UnicodeReader:
     """Handle buffering partial Unicode data"""
 
     def __init__(self, encoding, textmode=False):
-        self._encoding = encoding
-        self._textmode = textmode
-        self._partial = b''
+        if encoding and not textmode:
+            self._decoder = codecs.getincrementaldecoder(encoding)()
+        else:
+            self._decoder = None
 
-    def decode(self, data):
+    def decode(self, data, final=False):
         """Decode Unicode bytes when reading from binary sources"""
 
-        if self._encoding and not self._textmode:
-            data = self._partial + data
-            self._partial = b''
-
+        if self._decoder:
             try:
-                data = data.decode(self._encoding)
-            except UnicodeDecodeError as exc:
-                if exc.start > 0:
-                    # Avoid pylint false positive
-                    # pylint: disable=invalid-slice-index
-                    self._partial = data[exc.start:]
-                    data = data[:exc.start].decode(self._encoding)
-                elif exc.reason == 'unexpected end of data':
-                    self._partial = data
-                    data = ''
-                else:
-                    self.close()
-                    raise DisconnectError(DISC_PROTOCOL_ERROR,
-                                          'Unicode decode error')
+                data = self._decoder.decode(data, final)
+            except UnicodeDecodeError:
+                self.close()
+                raise DisconnectError(DISC_PROTOCOL_ERROR,
+                                      'Unicode decode error')
 
         return data
 
     def check_partial(self):
         """Check if there's partial Unicode data left at EOF"""
 
-        if self._partial:
-            self.close()
-            raise DisconnectError(DISC_PROTOCOL_ERROR, 'Unicode decode error')
+        self.decode(b'', True)
 
     def close(self):
         """Perform necessary cleanup on error (provided by derived classes)"""
@@ -83,14 +72,16 @@ class _UnicodeWriter:
     """Handle encoding Unicode data before writing it"""
 
     def __init__(self, encoding, textmode=False):
-        self._encoding = encoding
-        self._textmode = textmode
+        if encoding and not textmode:
+            self._encoder = codecs.getincrementalencoder(encoding)()
+        else:
+            self._encoder = None
 
     def encode(self, data):
         """Encode Unicode bytes when writing to binary targets"""
 
-        if self._encoding and not self._textmode:
-            data = data.encode(self._encoding)
+        if self._encoder:
+            data = self._encoder.encode(data)
 
         return data
 
@@ -137,6 +128,39 @@ class _FileReader(_UnicodeReader):
         self._file.close()
 
 
+class _AsyncFileReader(_FileReader):
+    """Forward data from an aiofile"""
+
+    def __init__(self, process, file, bufsize, datatype, encoding):
+        super().__init__(process, file, bufsize, datatype, encoding)
+
+        self._conn = process.channel.get_connection()
+
+    @asyncio.coroutine
+    def _feed(self):
+        """Feed file data"""
+
+        while not self._paused:
+            data = yield from self._file.read(self._bufsize)
+
+            if data:
+                self._process.feed_data(self.decode(data), self._datatype)
+            else:
+                self.check_partial()
+                self._process.feed_eof(self._datatype)
+                break
+
+    def feed(self):
+        """Start feeding file data"""
+
+        self._conn.create_task(self._feed())
+
+    def close(self):
+        """Stop forwarding data from the file"""
+
+        self._conn.create_task(self._file.close())
+
+
 class _FileWriter(_UnicodeWriter):
     """Forward data to a file"""
 
@@ -161,6 +185,25 @@ class _FileWriter(_UnicodeWriter):
         self._file.close()
 
 
+class _AsyncFileWriter(_FileWriter):
+    """Forward data to an aiofile"""
+
+    def __init__(self, process, file, encoding):
+        super().__init__(file, encoding)
+
+        self._conn = process.channel.get_connection()
+
+    def write(self, data):
+        """Write data to the file"""
+
+        self._conn.create_task(self._file.write(self.encode(data)))
+
+    def close(self):
+        """Stop forwarding data to the file"""
+
+        self._conn.create_task(self._file.close())
+
+
 class _PipeReader(_UnicodeReader, asyncio.Protocol):
     """Forward data from a pipe"""
 
@@ -536,6 +579,8 @@ class SSHProcess:
         else:
             if isinstance(source, str):
                 file = open(source, 'rb', buffering=bufsize)
+            elif isinstance(source, PurePath):
+                file = open(str(source), 'rb', buffering=bufsize)
             elif isinstance(source, int):
                 file = os.fdopen(source, 'rb', buffering=bufsize)
             elif isinstance(source, socket.socket):
@@ -543,7 +588,10 @@ class SSHProcess:
             else:
                 file = source
 
-            if _is_regular_file(file):
+            if hasattr(file, 'read') and asyncio.iscoroutinefunction(file.read):
+                reader = _AsyncFileReader(self, file, bufsize,
+                                          datatype, self._encoding)
+            elif _is_regular_file(file):
                 reader = _FileReader(self, file, bufsize,
                                      datatype, self._encoding)
             else:
@@ -584,6 +632,8 @@ class SSHProcess:
         else:
             if isinstance(target, str):
                 file = open(target, 'wb', buffering=bufsize)
+            elif isinstance(target, PurePath):
+                file = open(str(target), 'wb', buffering=bufsize)
             elif isinstance(target, int):
                 file = os.fdopen(target, 'wb', buffering=bufsize)
             elif isinstance(target, socket.socket):
@@ -591,7 +641,10 @@ class SSHProcess:
             else:
                 file = target
 
-            if _is_regular_file(file):
+            if hasattr(file, 'write') and \
+                    asyncio.iscoroutinefunction(file.write):
+                writer = _AsyncFileWriter(self, file, self._encoding)
+            elif _is_regular_file(file):
                 writer = _FileWriter(file, self._encoding)
             else:
                 if hasattr(target, 'buffer'):
@@ -775,7 +828,9 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
         recv_buf = self._recv_buf[datatype]
 
         if recv_buf and isinstance(recv_buf[-1], Exception):
-            recv_buf = recv_buf[:-1]
+            recv_buf, self._recv_buf[datatype] = recv_buf[:-1], recv_buf[-1:]
+        else:
+            self._recv_buf[datatype] = []
 
         buf = '' if self._encoding else b''
         return buf.join(recv_buf)
@@ -832,7 +887,8 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
                * A file object open for read
                * An `int` file descriptor open for read
                * A connected socket object
-               * A string containing the name of a file or device to open
+               * A string or :class:`PurePath <pathlib.PurePath>` containing
+                 the name of a file or device to open
                * `DEVNULL` to provide no input to standard input
                * `PIPE` to interactively write standard input
 
@@ -842,7 +898,8 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
                * A file object open for write
                * An `int` file descriptor open for write
                * A connected socket object
-               * A string containing the name of a file or device to open
+               * A string or :class:`PurePath <pathlib.PurePath>` containing
+                 the name of a file or device to open
                * `DEVNULL` to discard standard error output
                * `PIPE` to interactively read standard error output
 
@@ -903,6 +960,21 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
 
         yield from self.redirect(None, None, target, bufsize, send_eof)
 
+    def collect_output(self):
+        """Collect output from the process without blocking
+
+           This method returns a tuple of the output that the process
+           has written to stdout and stderr which has not yet been read.
+           It is intended to be called instead of read() by callers
+           that want to collect received data without blocking.
+
+           :returns: A tuple of output to stdout and stderr
+
+        """
+
+        return (self._collect_output(),
+                self._collect_output(EXTENDED_DATA_STDERR))
+
     # pylint: disable=redefined-builtin
     @asyncio.coroutine
     def communicate(self, input=None):
@@ -930,8 +1002,7 @@ class SSHClientProcess(SSHProcess, SSHClientStreamSession):
 
         yield from self._chan.wait_closed()
 
-        return (self._collect_output(),
-                self._collect_output(EXTENDED_DATA_STDERR))
+        return self.collect_output()
     # pylint: enable=redefined-builtin
 
     def change_terminal_size(self, width, height, pixwidth=0, pixheight=0):
@@ -1091,7 +1162,8 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
                * A file object open for write
                * An `int` file descriptor open for write
                * A connected socket object
-               * A string containing the name of a file or device to open
+               * A string or :class:`PurePath <pathlib.PurePath>` containing
+                 the name of a file or device to open
                * `DEVNULL` to discard standard error output
                * `PIPE` to interactively read standard error output
 
@@ -1101,7 +1173,8 @@ class SSHServerProcess(SSHProcess, SSHServerStreamSession):
                * A file object open for read
                * An `int` file descriptor open for read
                * A connected socket object
-               * A string containing the name of a file or device to open
+               * A string or :class:`PurePath <pathlib.PurePath>` containing
+                 the name of a file or device to open
                * `DEVNULL` to provide no input to standard input
                * `PIPE` to interactively write standard input
 
